package dunolint

  1. Overview
  2. Docs
On This Page
  1. Getters
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Dunolint_engine.Dune_project_contextSource

Context associated with a dune-project file.

When performing linting it may be useful to refer to context from the enclosing dune-project file. This type is used to hold the information we wish to access.

Sourcetype t

Each value of type t refers to context found in exactly one file. This type is non mutable.

create ~path ~original_contents parses information needed from a "dune-project" file, and extracts from it an immutable context required for linting.

Getters

A valid dune-project file will necessarily start with a valid lang dune stanza, however if the file is not valid, this may return None.

Sourcemodule Invalid_dune_project : sig ... end